@@ -104,8 +104,7 @@ meta_window_ensure_frame (MetaWindow *window)
                                                 frame->rect.x,
                                                 frame->rect.y,
 						frame->rect.width,
-						frame->rect.height,
-						frame->window->screen->number);
+						frame->rect.height);
 
   meta_verbose ("Frame for %s is 0x%lx\n", frame->window->desc, frame->xwindow);
   attrs.event_mask = EVENT_MASK;
@@ -267,7 +266,7 @@ meta_frame_get_flags (MetaFrame *frame)
   if (META_WINDOW_ALLOWS_VERTICAL_RESIZE (frame->window))
     flags |= META_FRAME_ALLOWS_VERTICAL_RESIZE;
 
-  if (frame->window->has_focus)
+  if (meta_window_appears_focused (frame->window))
     flags |= META_FRAME_HAS_FOCUS;
 
   if (frame->window->shaded)
